@@ -5,7 +5,7 @@ source ./.travis/lib-setup.sh
 setup::install
 
 # always build in-place so that Sphinx can find the modules
-python setup.py build_ext --inplace
+python setup.py build_ext --inplace $EXTRA_BUILDEXT
 BUILD_RES=$?
 
 if [ x"$KRB5_VER" = "xheimdal" ]; then
@@ -13,6 +13,11 @@ if [ x"$KRB5_VER" = "xheimdal" ]; then
     exit $BUILD_RES
 fi
 
+if [ "$TRAVIS_OS_NAME" == "windows" ]; then
+    # Windows can't run tests yet, so just exit
+    exit $BUILD_RES
+fi
+
 if [ $BUILD_RES -ne 0 ]; then
     # if the build failed, don't run the tests
     exit $BUILD_RES
